import base64
import os
import tempfile

import fitz
from math import atan2, degrees

from .llm import client, use_tools
from pprint import pprint as print

from habanero import Crossref
cr = Crossref()


def summarize_paper(text):
    system_message = ("Given a block of text, your task is to summarize the text into a concise paragraph. "
                      "Do not include any references or citations in the summary. "
                      "Do not speak to the user directly, just produce the summary of the text you are given.")

    user_message = "Summarize this text:\n\n" + text

    messages = [
        {
            "role": "system",
            "content": system_message
        },
        {
            "role": "user",
            "content": user_message
        }
    ]

    arguments = {"messages": messages, "model": "gpt-4o"}

    response = client.chat.completions.create(**arguments)
    return response.choices[0].message.content


def extract_doi(images, incorrect_doi_list=None):
    tools = [
        {
            "type": "function",
            "function": {
                "name": "store_doi",
                "description": "Store the DOI in the database",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "doi": {
                            "type": "string",
                            "description": "The DOI to store",
                        },
                    },
                    "required": ["doi"],
                },
            }
        },
        {
            "type": "function",
            "function": {
                "name": "keep_searching_for_doi",
                "description": "Keep searching for the DOI",
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            }
        }
    ]
    system_message = ("Given a block of text, your task is to extract the DOI from the text. "
                      "The DOI is a unique alphanumeric string that provides a permanent link to the location of an "
                      "online resource. It is often found in the header, footer, or metadata of a research paper. "
                      "If the DOI is not present in the text, please use the keep_searching_for_doi function. "
                      "If the DOI is found, please store it in the database for future reference by using the "
                      "store_doi function. "
                      "\nExample of DOI: '12.3456/nature123'. If the DOI is in the form of a URL, please extract the "
                      "DOI from the URL and store the DOI without the URL format. ")
    user_message_prefix = "Extract the DOI from this image"

    if incorrect_doi_list:
        user_message_prefix = (". The DOI is not any of these '" + ", ".join(incorrect_doi_list) +
                               "'. Extract the correct DOI from this image")

    for image in images:
        messages = [
            {
                "role": "system",
                "content": system_message
            },
            {
                "role": "user",
                "content": [
                            {
                              "type": "image_url",
                              "image_url": {
                                  "url": image,
                              }
                            },
                            {
                              "type": "text",
                              "text": user_message_prefix
                            }
                          ]
            }
        ]

        arguments = {"messages": messages, "tools": tools, "model": "gpt-4o"}

        retry = 0
        valid_calls = []
        while valid_calls == [] and retry < 3:
            if retry > 0:
                print("Retrying...")
            chat_response = client.chat.completions.create(**arguments)
            if chat_response.choices[0].message.tool_calls:
                valid_calls = use_tools(chat_response, arguments, call_functions=False)
                if valid_calls:
                    for call in valid_calls:
                        if call["name"] == "store_doi":
                            return call["parameters"]["doi"]
            retry += 1
    return None


def create_cleaned_text(images):

    figure_present_system_message = ("Read the contents of the provided scan of a page from a research paper. "
                                     "Record the number of figures and tables that are present on the page.")

    figure_present_tools = [
        {
            "type": "function",
            "function": {
                "name": "store_figure_table_count",
                "description": "Store the count of figures and tables",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "figure_count": {
                            "type": "integer",
                            "description": "The number of figures on the page",
                        },
                        "table_count": {
                            "type": "integer",
                            "description": "The number of tables on the page",
                        }
                    },
                    "required": ["figure_count", "table_count"],
                },
            }
        }
    ]

    figure_system_message = ("Read the contents of the provided scan of a page from a research paper. "
                             "For each figure and table in the paper include a '**Figure/Table Description:**' "
                             "section that you will author. This section should include a description of what is "
                             "being communicated in the figure or table based on your best impression as well as "
                             "all text that is found within that figure. Make sure to lay out figure or table text "
                             "in a manner that best communicates the intent of the author with the rest of the "
                             "output. You should always include a '**Figure/Table Description:**' for every "
                             "figure and table you see in the scan. Once you have written out the text for "
                             "all figures write **FIGURES_AND_TABLES_COMPLETE**")

    body_system_message = ("Read the contents of the provided scan of a page from a research paper. "
                           "Convert the tex that is in the main body of the paper to raw text. "
                           "Do not include any tables. Do not include any figures. Do not include any footnotes. "
                           "Do not include any reference sections. Once you have written out the text in the main "
                           "body of the paper write  **PAGE_COMPLETE** and stop.")

    first_page_system_message = ("Read the contents of the provided scan of a page from a research paper. "
                                 "Convert the text of paper to raw text. Skip the title, authors, headers, footers, "
                                 "legalese, copyrights, and references. Include the abstract and any other "
                                 "introductory text as well as the main body of the paper. Once you have written out "
                                 "the text in the main body of the paper write **PAGE_COMPLETE** and stop.")

    cleaned_text = ""

    for i, image in enumerate(images):
        print("Processing page " + str(i + 1))

        cleaned_text += "\n\n**Start of Page " + str(i + 1) + "**\n\n"

        messages = [
            {
                "role": "system",
                "content": body_system_message
            },
            {
                "role": "user",
                "content": [
                            {
                              "type": "image_url",
                              "image_url": {
                                  "url": image,
                              }
                            }
                          ]
            }
        ]

        if i == 0:
            messages[0]["content"] = first_page_system_message

        arguments = {"messages": messages, "model": "gpt-4o", "temperature": 0.2}

        chat_response = client.chat.completions.create(**arguments)

        cleaned_text += chat_response.choices[0].message.content.replace("**PAGE_COMPLETE**", "")

        messages = [
            {
                "role": "system",
                "content": figure_present_system_message
            },
            {
                "role": "user",
                "content": [
                            {
                              "type": "image_url",
                              "image_url": {
                                  "url": image,
                              }
                            }
                          ]
            }
        ]

        arguments = {"messages": messages, "tools": figure_present_tools, "model": "gpt-4o", "temperature": 0.2,
                     "tool_choice": {"type": "function", "function": {"name": "store_figure_table_count"}}}

        retry = 0
        table_figure_count = -1
        while table_figure_count < 0 and retry < 3:
            if retry > 0:
                print("Retrying...")
            chat_response = client.chat.completions.create(**arguments)
            if chat_response.choices[0].message.tool_calls:
                valid_calls = use_tools(chat_response, arguments, call_functions=False)
                if valid_calls:
                    for call in valid_calls:
                        if call["name"] == "store_figure_table_count":
                            try:
                                table_figure_count = (int(call["parameters"]["figure_count"]) +
                                                      int(call["parameters"]["table_count"]))
                            except Exception as e:
                                table_figure_count = -1
            retry += 1

        if table_figure_count < 0 or table_figure_count > 0:
            messages = [
                {
                    "role": "system",
                    "content": figure_system_message
                },
                {
                    "role": "user",
                    "content": [
                                {
                                  "type": "image_url",
                                  "image_url": {
                                      "url": image,
                                  }
                                }
                              ]
                }
            ]

            arguments = {"messages": messages, "model": "gpt-4o", "temperature": 0.2}

            chat_response = client.chat.completions.create(**arguments)

            cleaned_text += chat_response.choices[0].message.content.replace("**FIGURES_AND_TABLES_COMPLETE**", "")

        cleaned_text += "\n\n**End of Page " + str(i + 1) + "**\n\n"

    return cleaned_text


def confirm_doi(title, images):
    system_message = ("Read the contents of the provided scan of a page from a research paper. "
                      "Extract the title of the paper from the text")
    tools = [
        {
            "type": "function",
            "function": {
                "name": "store_title",
                "description": "Store the title in the database",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "title": {
                            "type": "string",
                            "description": "The title to store",
                        },
                    },
                    "required": ["title"],
                },
            }
        },
    ]
    arguments = {"messages": [
        {
            "role": "system",
            "content": system_message
        },
        {
            "role": "user",
            "content": [
                        {
                          "type": "image_url",
                          "image_url": {
                              "url": images[0],
                          }
                        }
                      ]
        }
    ], "tools": tools, "model": "gpt-4o", "temperature": 0.2,
        "tool_choice": {"type": "function", "function": {"name": "store_title"}}}

    retry = 0
    title_found = False
    while not title_found and retry < 3:
        chat_response = client.chat.completions.create(**arguments)
        if chat_response.choices[0].message.tool_calls:
            valid_calls = use_tools(chat_response, arguments, call_functions=False)
            if valid_calls:
                for call in valid_calls:
                    if call["name"] == "store_title":
                        stored_title = call["parameters"]["title"]
                        title_found = True
        retry += 1

    if not title_found:
        return False

    system_message = ("Are these titles likely to be the same?")

    tools = [
        {
            "type": "function",
            "function": {
                "name": "store_title_similar",
                "description": "Store the title similarity in the database",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "titles_similar": {
                            "type": "boolean",
                            "description": "Are the titles similar?",
                        },
                    },
                    "required": ["titles_similar"],
                },
            }
        },
    ]

    arguments = {"messages": [{"role": "system", "content": system_message},
                              {"role": "user", "content": "Title 1: " + title + "\nTitle 2: " + stored_title}],
                 "model": "gpt-4o", "temperature": 0.2, "tools": tools,
                 "tool_choice": {"type": "function", "function": {"name": "store_title_similar"}}}

    print("Checking title similarity... "
          "Title 1: " + title + "\nTitle 2: " + stored_title)

    retry = 0
    title_similarity = None
    while title_similarity is None and retry < 3:
        chat_response = client.chat.completions.create(**arguments)
        if chat_response.choices[0].message.tool_calls:
            valid_calls = use_tools(chat_response, arguments, call_functions=False)
            if valid_calls:
                for call in valid_calls:
                    if call["name"] == "store_title_similar":
                        title_similarity = call["parameters"]["titles_similar"]
        retry += 1

    return title_similarity


def rotate_pdf_pages(pdf_path):
    doc = fitz.open(pdf_path)
    for page in doc:
        text_blocks = page.get_text("dict")["blocks"]
        total_weight = 0
        weighted_sum_angles = 0

        # Collect angles of text blocks
        for block in text_blocks:
            if block["type"] == 0:  # Text block
                for line in block['lines']:
                    dir_vector = line['dir']
                    for span in line['spans']:
                        angle = atan2(dir_vector[1], dir_vector[0])
                        text_length = len(span['text'])
                        weighted_sum_angles += degrees(angle) * text_length
                        total_weight += text_length

        weighted_average_angle = weighted_sum_angles / total_weight if total_weight else 0

        # Calculate the average angle if angles were detected
        if total_weight > 0:
            average_angle = weighted_average_angle

            # Determine the rotation needed to align text upright
            # We assume that text should be as close to 0 degrees as possible
            # This might need adjustments for specific use cases
            if average_angle != 0:
                # Normalize the average angle to the nearest multiple of 90
                # This is a simplistic approach; more sophisticated logic may be needed
                normalized_angle = 360 - (round(average_angle / 90) * 90)
                page.set_rotation(normalized_angle)
    modified_pdf_path = pdf_path.replace(".pdf", "_rotated.pdf")
    doc.save(modified_pdf_path)
    doc.close()
    os.rename(modified_pdf_path, pdf_path)
    return


def gather_metadata(pages):
    retry = 3
    old_doi_list = None
    found_doi = False
    while retry > 0 and not found_doi:
        doi = extract_doi(pages, incorrect_doi_list=old_doi_list)
        try:
            crossref_data = cr.works(ids=doi)
            title = crossref_data["message"]["title"][0]
            print("Title: " + title)
            found_doi = confirm_doi(title, pages)
            print("Found DOI: " + str(found_doi))
            if not found_doi:
                if not old_doi_list:
                    old_doi_list = [doi]
                else:
                    old_doi_list.append(doi)
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            retry -= 1
    if not found_doi:
        raise Exception("DOI not found")

    metadata = crossref_data["message"]

    dois = []
    for ref in metadata["reference"]:
        if "DOI" in ref:
            dois.append(ref["DOI"])

    references = []
    ref_number = 0
    try:
        # Retrieve metadata for multiple DOIs at once
        res = cr.works(ids=dois)
        for item in res:
            ref_number += 1
            data = item['message']
            # Format the reference string based on available fields
            author_str = ', '.join([author['given'] + ' ' + author['family'] for author in data.get('author', []) if
                                    'given' in author and 'family' in author])
            title_str = data.get('title', [''])[0] if data.get('title', None) else ''
            journal_str = data.get('container-title', [''])[0] if data.get('container-title', None) else ''
            volume_str = data.get('volume', '')
            page_str = data.get('page', '')
            year_str = str(data['issued']['date-parts'][0][0]) if data.get('issued', None) else ''
            doi_str = data.get('DOI', '')
            reference_str = f"{ref_number}. {author_str}. {title_str}. {journal_str}, {volume_str}, {page_str}, " \
                            f"{year_str}. DOI: {doi_str}"
            references.append(reference_str.strip())
    except Exception as e:
        references.append(f"Details not retrieved due to error: {e}")
    return references, metadata


def process_paper(pdf_path):
    temp_folder = tempfile.mkdtemp()
    os.makedirs(temp_folder, exist_ok=True)
    # Open the PDF file
    rotate_pdf_pages(pdf_path)
    doc = fitz.open(pdf_path)
    # use fitz to create a clear image of each page make sure to use a high DPI
    image_list = []
    for i in range(len(doc)):
        page = doc[i]
        mat = fitz.Matrix(200 / 72, 200 / 72)
        image = page.get_pixmap(matrix=mat)
        #image = page.get_pixmap()
        image_bytes = image.tobytes()
        image_list.append(image_bytes)

    # # write the first image to a file to test
    # with open("page0.png", "wb") as f:
    #     f.write(image_list[0])

    page_images = []
    for image_bytes in image_list:
        base64_image = base64.b64encode(image_bytes).decode("utf-8")
        page_image = "data:image/png;base64," + base64_image
        page_images.append(page_image)

    output = {}
    references, metadata = gather_metadata(page_images)
    title = metadata["title"][0]
    output["page_images"] = page_images
    output["cleaned_text"] = (title + "\n\n\n" +
                              create_cleaned_text(page_images) + "\n\n## REFERENCES\n" + "\n".join(references))
    output["metadata"] = metadata

    output["summary"] = summarize_paper(output["cleaned_text"])

    return output



from openai import OpenAI
import tiktoken
import json
import os
import threading
import traceback


class GenericClientWrapper:
    def __init__(self, client, stop_event):
        self._client = client
        self._stop_event = stop_event

    def __getattr__(self, name):
        attr = getattr(self._client, name)
        if self._stop_event.is_set():
            quit(0)
        if callable(attr):
            def wrapper(*args, **kwargs):
                if self._stop_event.is_set():
                    quit(0)
                try:
                    return attr(*args, **kwargs)
                except Exception as e:
                    print(f"Request failed: {e}")
                    return None
            return wrapper
        return attr


base_key_path = os.path.join(os.path.expanduser("~"), "Documents", "ScienceAI")
target_key = os.path.join(base_key_path, "scienceai-keys.json")
if not os.path.exists(target_key):
    new_key = input("Please enter OpenAI key: ")
    with open(target_key, "w") as file:
        json.dump({"openai": new_key}, file)
with open(target_key, "r") as file:
    key_list = json.load(file)
openai_key = key_list.get("openai")
if openai_key is None:
    raise Exception("Open AI key not in scienceai-keys.json")

try:
    __client = OpenAI(api_key=openai_key)


    def is_api_key_valid():
        response = __client.chat.completions.create(model="gpt-4o", messages=[{"role": "user", "content": "Hello, "}],
                                                      max_tokens=1)


    is_api_key_valid()
except Exception as e:
    delete = input("Error creating OpenAI client - delete saved key? (Y/n): ")
    if delete.lower() == "y":
        os.remove(target_key)
        new_key = input("Please enter OpenAI key: ")
        with open(target_key, "w") as file:
            json.dump({"openai": new_key}, file)
        __client = OpenAI(api_key=new_key)
    else:
        raise e

enc = tiktoken.encoding_for_model("gpt-4")

stop_event = threading.Event()

client = GenericClientWrapper(__client, stop_event)


def update_stop_event(event):
    client._stop_event = event


def check_tool(function_name, parameters, tools):
    function_names = []
    for tool in tools:
        function_names.append(tool["function"]["name"])
    if function_name not in function_names:
        return {"role": "system",
                "content": "ERROR: The function name is not in the list of functions provided to you. "
                           "Please try again."}
    if isinstance(parameters, str):
        try:
            parameters = json.loads(parameters)
        except json.JSONDecodeError:
            return {"role": "system",
                    "content": "ERROR: You're function call did not parse as valid JSON. Please try again."}
    if isinstance(parameters, dict):
        for key in parameters:
            if key not in tools[function_names.index(function_name)]["function"]["parameters"]["properties"]:
                return {"role": "system",
                        "content": "ERROR: The parameter '" + key +
                                   "' is not in the list of parameters provided to you. Please try again."}
    return None


def trim_history(history, token_limit):
    for i in range(len(history)):
        if len(enc.encode(str(history))) > token_limit:
            history.pop(1)
        else:
            return history


def use_tools(chat_response, arguments, function_dict={}, call_functions=True, pre_tool_call=False):
    """
    Use the tools specified in the tool_calls dict.
    :param chat_response:
    :type chat_response: dict
    :param arguments:
    :type arguments: dict
    :param function_dict:
    :type function_dict: dict
    :param call_functions:
    :type call_functions: bool
    :param pre_tool_call:
    :type pre_tool_call: bool
    :return: list of history items
    :rtype: list
    """
    if isinstance(chat_response, dict):
        tool_calls = chat_response['tool_calls']
        content = chat_response['content']
    else:
        tool_calls = chat_response.choices[0].message.tool_calls
        content = chat_response.choices[0].message.content
    tools = arguments['tools']
    tool_calls_list = []
    if tool_calls:
        for tool_call in tool_calls:
            if isinstance(tool_call, dict):
                tool_calls_list.append({
                    "function": {
                        "arguments": tool_call['function']['arguments'],
                        "name": tool_call['function']['name'],
                    },
                    "id": tool_call['id'], "type": 'function'
                })
            else:
                tool_calls_list.append({'function': {
                    "arguments": tool_call.function.arguments,
                    "name": tool_call.function.name,
                }, 'id': tool_call.id, 'type': 'function'})
    if call_functions:
        if tool_calls_list:
            new_history = [{"content": content, "role": "assistant", "tool_calls": tool_calls_list}]
        else:
            new_history = [{"content": content, "role": "assistant"}]
    if pre_tool_call:
        return new_history
    tool_results = []
    tool_errors = []
    valid_calls = []
    for tool_call in tool_calls_list:
        function_name = tool_call['function']['name']
        tool_schema = None
        for schema in tools:
            if schema['function']['name'] == function_name:
                tool_schema = schema
        results, errors = use_tool(tool_call['function'], tool_call['id'], tool_schema, function_dict=function_dict, call_functions=call_functions)
        if call_functions:
            tool_results += results
            tool_errors += errors
        else:
            valid_calls += results
    if call_functions:
        return new_history + tool_results+tool_errors
    else:
        return valid_calls


def use_tool(tool_call, tool_id, tool_schema, function_dict={}, call_functions=True):
    """
    Use the tool specified in the tool_call dict.
    :param tool_call:
    :type tool_call: dict
    :param tool_id:
    :type tool_id: str
    :param tool_schema:
    :type tool_schema: dict
    :param function_dict:
    :type function_dict: dict
    :param call_functions:
    :type call_functions: bool
    :return: results and errors
    :rtype: tuple
    """
    function_name = tool_call['name']
    results = []
    errors = []
    missing_function = False
    if call_functions:
        try:
            called_function = function_dict[function_name]
        except KeyError:
            missing_function = True
            errors.append({"content": "ERROR", "role": "tool",
                           "name": function_name})
            errors.append({"content": "Only use a valid function in your "
                                      "function list.", "role": "system"})
    if not missing_function:
        try:
            arguments = json.loads(tool_call['arguments'])
            try:
                if call_functions:
                    result = called_function(**arguments)
                    results.append({"role": "tool", "name": function_name, "content": str(result), "tool_call_id": tool_id})
                else:
                    results.append({"role": "function_call", "name": function_name, "parameters": arguments})
            except Exception as e:
                error_str = ("Error calling " + function_name + " function with passed arguments " +
                             str(arguments) + " : " + traceback.format_exc() + " \n " + str(e))
                errors.append({"content": error_str, "role": "tool", "tool_call_id": tool_id,
                               "name": function_name})
        except json.decoder.JSONDecodeError:
            required_arguments = tool_schema['function']['parameters']['required']
            if tool_call['arguments'] == "":
                new_history_item = {"content": "You're function call did not "
                                               "include any arguments. Please try again with the "
                                               "correct arguments: " + str(required_arguments),
                                    "role": "system"}
            else:
                new_history_item = {"content": "You're function call did not parse as valid JSON. "
                                               "Please try again", "role": "system"}
            errors.append({"content": "ERROR", "role": "tool", "name": function_name, "tool_call_id": tool_id})
            errors.append(new_history_item)
    return results, errors



import json
import os
from .llm import client, use_tools


data_types = None
data_types_file = "data_types.json"

data_types_docs = None
data_types_docs_file = "data_types_docs.json"

path_to_app = os.path.dirname(os.path.abspath(__file__))
data_types_file = os.path.join(path_to_app, data_types_file)
data_types_docs_file = os.path.join(path_to_app, data_types_docs_file)


def load_json_file(filename):
    """ Load and return the data from a JSON file. """
    with open(filename, 'r') as file:
        return json.load(file)


data_types = load_json_file(data_types_file)
data_types_docs = load_json_file(data_types_docs_file)


def validate_data_type_spec(spec, force_type=None):
    if force_type:
        if force_type not in data_types:
            return False, f"Data type '{force_type}' not found in data types."
        properties = data_types[force_type]['spec']
        specification = spec
    else:
        if 'type' not in spec:
            return False, "Key 'type' not found in specification."
        if spec['type'] not in data_types:
            return False, f"Data type '{spec['type']}' not found in data types."
        properties = data_types[spec['type']]['spec']
        specification = spec.copy()
        del specification['type']
    for key, value in properties.items():
        if key not in specification:
            return False, f"Key '{key}' not found in specification."
        if value['type'] == 'string':
            if not isinstance(specification[key], str):
                return False, f"Value for '{key}' is not a string."
        elif value['type'] == 'boolean':
            if not isinstance(specification[key], bool):
                return False, f"Value for '{key}' is not a boolean."
        elif value['type'] == 'number':
            if not isinstance(specification[key], int) and not isinstance(specification[key], float):
                return False, f"Value for '{key}' is not a number."
        elif value['type'] == 'object':
            try:
                for k, v in json.loads(specification[key]):
                    if k in value['keys']:
                        if value['keys'][k] == 'string':
                            if not isinstance(v, str):
                                return False, f"Value for '{k}' is not a string."
                        elif value['keys'][k] == 'boolean':
                            if not isinstance(v, bool):
                                return False, f"Value for '{k}' is not a boolean."
                        else:
                            return False, f"Invalid data type '{value['keys'][k]}' for key '{k}' in object spec."
            except json.JSONDecodeError:
                return False, f"Value for '{key}' is not a valid JSON object."
        elif value['type'] == 'array':
            if not isinstance(specification[key], list):
                return False, f"Value for '{key}' is not an array."
            for item in specification[key]:
                if not isinstance(item, str):
                    return False, f"Item in array '{key}' is not a string."
    if len(specification) != len(properties):
        return False, "specification has extra keys."
    return True, "specification is valid."


def generate_description(data_type_key, data_type_value, all_keys):
    """ Generate a description for each data type. """
    message = f"Data Type: {data_type_key}\n"
    message += f"Description: {data_type_value['spec_description']}\n"
    message += "Specifications Required:\n"
    properties = {}
    for spec_key, spec_value in data_type_value['spec'].items():
        message += f"  - {spec_key} ({spec_value['type']}): {spec_value['description']}\n"
        properties[spec_key] = {"type": spec_value['type'], "description": spec_value['description']}
        if spec_value['type'] == 'array':
            properties[spec_key]["items"] = {"type": "string"}
    tools = [
        {
            "type": "function",
            "function": {
                "name": "generate_example_data",
                "description": "creates example data for a given data type",
                "parameters": {
                    "type": "object",
                    "properties": properties,
                    "required": list(properties.keys())
                }
            }
        }
    ]

    system_message = ("You are a helpful AI assistant. You have been asked to generate example data for data types. "
                      "These data types are used to extract data from research papers. Each aspect of the example data "
                      "must be creative and scientific. You must follow the specifications provided for each "
                      "data type. Don't repeat the same example data for the same data type. Be creative! "
                      "If there is a boolean value, make sure to vary it between examples. "
                      "(Not every data type should be required: true!)"
                      "\n\n\n Keep in mind that the full list of data types is: " + ", ".join(all_keys) + "\n\n\n"
                      "Be sure to generate example data for the following data type: " + data_type_key + "\n\n"
                      "Make sure your example data is not too similar to other data types.")

    user_message_prefix = "Generate made up example data for this data type that is creative and scientific:\n\n"

    messages = [
        {
            "role": "system",
            "content": system_message
        },
        {
            "role": "user",
            "content": user_message_prefix + message
        }
    ]

    arguments = {"messages": messages, "tools": tools, "model": "gpt-4o",
                 "tool_choice": {"type": "function", "function": {"name": "generate_example_data"}}}

    examples = []
    retry = 0
    valid_calls = []
    output_dictionary = None
    while len(examples) < 3 and retry < 3:
        while valid_calls == [] and retry < 3:
            if retry > 0:
                print("Retrying...")
            chat_response = client.chat.completions.create(**arguments)
            if chat_response.choices[0].message.tool_calls:
                valid_calls = use_tools(chat_response, arguments, call_functions=False)
                if valid_calls:
                    for call in valid_calls:
                        if call["name"] == "generate_example_data":
                            output_dictionary = call["parameters"]
                            valid, error_message = validate_data_type_spec(output_dictionary, force_type=data_type_key)
                            if not valid:
                                print(error_message, output_dictionary)
                                valid_calls = []
                            else:
                                output_dictionary["type"] = data_type_key
            else:
                print("No tool calls used")
            retry += 1
        if valid_calls and output_dictionary:
            retry = 0
            messages += [{"role": "function", "name": valid_calls[0]["name"], "content": "valid"},
                                       {"role": "user", "content": user_message_prefix + message}]
            examples.append(output_dictionary)
            valid_calls = []

    return message, examples


for data_type in data_types:
    if data_type not in data_types_docs:
        data_types_docs[data_type] = {}
        data_types_docs[data_type]["description"], data_types_docs[data_type]["examples"] = (
            generate_description(data_type, data_types[data_type], list(data_types.keys())))
        with open(data_types_docs_file, 'w') as file:
            json.dump(data_types_docs, file, indent=2)


def generate_schema(corpus, goal=None, retries=5):
    tools = [
        {
            "type": "function",
            "function": {
                "name": "generate_analysis_schema",
                "description": "creates example data for a given data type",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "schema": {"type": "string",
                                   "description": "A JSON string representing the schema for the analysis which should "
                                                  "be an array of objects with a 'type' key and other keys as "
                                                  "specified in the data types."},
                    },
                    "required": ["schema"]
                }
            }
        }
    ]

    system_message = ("You are an expert Data Scientist. You have been asked to generate the schema which will be used "
                      "to extract data from research papers. The schema must be creative and scientific. "
                      "You must follow the specifications provided for each data type.\n\nSpecifications for Schema:\n")

    for key, value in data_types_docs.items():
        examples = []
        for example in value["examples"]:
            examples.append(str(example))
        system_message += value["description"] + "\n" + key + " Examples:\n" + "\n".join(examples) + "\n\n"

    system_message += "\n\n\n"

    if goal:
        preamble = "Read the corpus and generate the schema with data types that will best address the following: "
        system_message += preamble + goal
        if not system_message.endswith("."):
            system_message += "."
    else:
        system_message += "Read the corpus and generate the schema with data types that will best sum up the corpus. "
    system_message += "\n\n IMPORTANT: basic bibliographic information has already been extracted from the corpus so "
    system_message += "you should not include that in the schema, so no need to ask for author names, titles, etc. "
    system_message += "Also, the schema will be used to extract data from "
    system_message += "each paper in the corpus SEPARATELY. Make sure the schema is general enough to apply to all "
    system_message += "papers in the corpus. Do not ask for ANY information that sums up the corpus as a whole."

    user_message = "Generate an analysis schema for this corpus"

    if goal:
        user_message += " that will best address the following goal - " + goal

    user_message += ":\n\n" + corpus + "\n"

    messages = [
        {
            "role": "system",
            "content": system_message
        },
        {
            "role": "user",
            "content": user_message
        }
    ]

    arguments = {"messages": messages, "tools": tools, "model": "gpt-4o",
                 "tool_choice": {"type": "function", "function": {"name": "generate_analysis_schema"}}}

    retry = 0
    valid_calls = []
    output_dictionary = None
    while valid_calls == [] and retry < retries:
        if retry > 0:
            print("Retrying...")
        chat_response = client.chat.completions.create(**arguments)
        if chat_response.choices[0].message.tool_calls:
            valid_calls = use_tools(chat_response, arguments, call_functions=False)
            if valid_calls:
                for call in valid_calls:
                    if call["name"] == "generate_analysis_schema":
                        output_dictionary = call["parameters"]
                        if "schema" in output_dictionary:
                            try:
                                output_dictionary = json.loads(output_dictionary["schema"])
                                for data_type in output_dictionary:
                                    valid, error_message = validate_data_type_spec(data_type)
                                    if not valid:
                                        print(error_message, output_dictionary)
                                        valid_calls = []
                            except json.JSONDecodeError:
                                print("Schema is not a valid JSON string:", output_dictionary["schema"])
                                valid_calls = []
        else:
            print("No tool calls used")
        retry += 1
        if valid_calls and output_dictionary:
            return output_dictionary
    return None


def schema_to_tool(schema):

    selected_data_types = {}
    required = []

    for data_type_requested in schema:
        data_type_def = data_types[data_type_requested["type"]]["tool"]
        if data_type_def["mode"] == "prefix":
            required_keys = list(data_type_def.keys())
            required_keys.remove("mode")
            name = data_type_requested["name"].replace(" ", "_")
            for key in required_keys:
                if data_type_requested["required"]:
                    required.append(name + "_" + key)
                new_type = data_type_def[key]["type"]
                new_description = data_type_def[key]["description"]
                for spec_key in data_types[data_type_requested["type"]]["spec"]:
                    if spec_key == 'name':
                        new_description = new_description.replace("NAME", name)
                    else:
                        new_description = new_description.replace(spec_key.upper(), str(data_type_requested[spec_key]))
                selected_data_types[name + "_" + key] = {"type": new_type, "description": new_description}
            selected_data_types[name + "_successfully_extracted"] = \
                {"type": "boolean", "description": "Was the data for "+name+" successfully extracted?"}
            required.append(name + "_successfully_extracted")
        if data_type_def["mode"] == "array":
            name = data_type_requested["name"].replace(" ", "_")
            if data_type_requested["required"]:
                required.append(name)
            new_description = data_type_def["description"]
            for spec_key in data_types[data_type_requested["type"]]["spec"]:
                if spec_key == 'name':
                    new_description = new_description.replace("NAME", name)
                else:
                    new_description = new_description.replace(spec_key.upper(), str(data_type_requested[spec_key]))

            item_properties = {}
            for key in data_type_def["keys"]:
                new_type = data_type_def["keys"][key]["type"]
                new_description = data_type_def["keys"][key]["description"]
                for spec_key in data_types[data_type_requested["type"]]["spec"]:
                    if spec_key == 'name':
                        new_description = new_description.replace("NAME", name)
                    else:
                        new_description = new_description.replace(spec_key.upper(), str(data_type_requested[spec_key]))
                item_properties[key] = {"type": new_type, "description": new_description}

            selected_data_types[name] = {"type": "array", "description": new_description,
                                         "items": {"type": "object", "properties": item_properties}}

            selected_data_types[name + "_successfully_extracted"] = \
                {"type": "boolean", "description": "Was the data for "+name+" successfully extracted?"}
            required.append(name + "_successfully_extracted")

    required = list(set(required))

    tools = [
        {
            "type": "function",
            "function": {
                "name": "extract_data",
                "description": "extracts data from a research paper using a schema",
                "parameters": {
                    "type": "object",
                    "properties": selected_data_types,
                    "required": required
                }
            }
        }
    ]
    return tools


def reflect_on_data_extraction(extraction, corpus, retries=3):

        system_message = "You are a careful data analyst. Reflect on the data extraction process."

        user_message = ("Was this data correctly extracted from this research paper? Explain why or why not. "
                        "\n\nData Extracted:\n\n") + extraction + "\n\nResearch Paper:" + corpus

        messages = [
            {
                "role": "system",
                "content": system_message
            },
            {
                "role": "user",
                "content": user_message
            }
        ]

        arguments = {"messages": messages, "model": "gpt-4o"}

        chat_response = client.chat.completions.create(**arguments)

        messages.append({"role": "assistant", "content": chat_response.choices[0].message.content})

        tools = [
            {
                "type": "function",
                "function": {
                    "name": "check_extracted_data",
                    "description": "Logs if the extracted data was 100% perfectly extracted.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "valid": {"type": "boolean", "description": "Whether the extracted data was "
                                                                        "100% perfectly extracted. Do not "
                                                                        "set this to true if there were any "
                                                                        "errors in the extraction or even"
                                                                        "minor mistakes."}
                        }
                    },
                    "required": ["valid"]
                }
            }
        ]

        arguments = {"messages": messages, "model": "gpt-4o", "tools": tools,
                     "tool_choice": {"type": "function", "function": {"name": "check_extracted_data"}}}

        retry = 0
        valid_calls = []
        while valid_calls == [] and retry < retries:
            if retry > 0:
                print("Retrying...")
            chat_response = client.chat.completions.create(**arguments)
            if chat_response.choices[0].message.tool_calls:
                valid_calls = use_tools(chat_response, arguments, call_functions=False)
                if valid_calls:
                    for call in valid_calls:
                        if call["name"] == "check_extracted_data":
                            return call["parameters"]["valid"]
            retry += 1
        return False


def remove_failed_data(data):
    successfull_keys = []
    for key, value in data.items():
        if key.endswith("_successfully_extracted"):
            if value:
                successfull_keys.append(key.replace("_successfully_extracted", ""))
    new_data = {}
    for key in successfull_keys:
        for k, v in data.items():
            if k.startswith(key) and not k.endswith("_successfully_extracted"):
                new_data[k] = v
    return new_data


def extract_data(tools, corpus, retries=5):

    system_message = "You are an careful data analyst. Dutifully find the data in the provided research paper."

    user_message = "Extract the requested data using the extract_data tool:\n\n" + corpus + "\n"

    messages = [
        {
            "role": "system",
            "content": system_message
        },
        {
            "role": "user",
            "content": user_message
        }
    ]

    arguments = {"messages": messages, "tools": tools, "model": "gpt-4o",
                 "tool_choice": {"type": "function", "function": {"name": "extract_data"}}}

    retry = 0
    valid_calls = []
    output_dictionary = None
    while valid_calls == [] and retry < retries:
        if retry > 0:
            print("Retrying...")
        chat_response = client.chat.completions.create(**arguments)
        if chat_response.choices[0].message.tool_calls:
            valid_calls = use_tools(chat_response, arguments, call_functions=False)
            if valid_calls:
                for call in valid_calls:
                    if call["name"] == "extract_data":
                        output_dictionary = call["parameters"]
                        output_dictionary = remove_failed_data(output_dictionary)
                        check = reflect_on_data_extraction(str(output_dictionary), corpus)
                        if not check:
                            print("Data extraction failed. Retrying...")
                            valid_calls = []
        else:
            print("No tool calls used")
        retry += 1
        if output_dictionary:
            return output_dictionary
    return None




from datetime import datetime
from scienceai.analyst import Analyst
from .database_manager import DatabaseManager
from .llm import client, use_tools
import os


path_to_app = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(path_to_app, "principle_investigator_base_prompt.txt"), "r") as file:
    system_message = file.read()


# PI Module
class PrincipalInvestigator:
    def __init__(self, dbr: DatabaseManager):
        self.db = dbr
        self.analysts = []
        analysts_db = dbr.get_all_analysts()
        for analyst_dict in analysts_db:
            self.analysts.append(Analyst(dbr, analyst_dict=analyst_dict))
        chat_db = self.db.get_database_chat()
        self.tool_callables = {
            "delegate_research": self.delegate_research
        }
        self.tools = [self.delegate_research(None, None, return_tool=True)]
        self.system_message = system_message
        first_message = ("Hello, I am ScienceAI. I first need to make sure all your papers are loaded into the system "
                         "before I can help you. I will let you know when I am ready to answer your questions. "
                         "This may take a long time if you uploaded many papers.")
        second_message = "All papers have been loaded into the system."
        defaults = [first_message, second_message]
        self.db.remove_old_default_messages(defaults)
        if len(chat_db) > 0:
            last_chat = chat_db[-1]
            if last_chat["content"] == first_message:
                self.db.update_last_chat("Pending")
                self.db.ingest_papers()
                self.db.process_all_papers()
                self.db.update_last_chat("Processed")
                second = {"content": second_message, "role": "system", "status": "Pending",
                          "time": datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')}
                self.db.add_chat(second)
                self.db.update_last_chat("Processed")
            elif last_chat["content"] == second_message:
                self.db.update_last_chat("Processed")
            else:
                if last_chat["status"] == "Pending":
                    if last_chat.get("tool_calls"):
                        self.finish_tool_calls(last_chat)
                    elif last_chat["role"] == "user":
                        self.process_message(last_chat["content"], last_chat["role"], last_chat["status"], last_chat["time"],
                                             store_message=False)
        else:
            first = {"content": first_message, "role": "system", "status": "Pending",
                     "time": datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')}
            self.db.add_chat(first)
            self.db.ingest_papers()
            self.db.process_all_papers()
            self.db.update_last_chat("Processed")
            second = {"content": second_message, "role": "system", "status": "Pending",
                      "time": datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')}
            self.db.add_chat(second)
            self.db.update_last_chat("Processed")
        self.db.update_last_chat("Processed")

    def delegate_research(self, name, question, return_tool=False):
        if return_tool:
            return {
                "type": "function",
                "function": {
                    "name": "delegate_research",
                    "description": "Delegates a specific research question pertaining to the "
                                   "uploaded database of research papers to a new Analyst Agent",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string",
                                "description": "Provide a name for this new analyst"
                            },
                            "question": {
                                "type": "string",
                                "description": "The specific sub-research question to be answered by the analyst"
                            }
                        }
                    },
                    "required": ["name", "question"],
                }
            }

        new_analyst = None
        if question is None:
            raise Exception("ERROR: Please provide a question for the analyst to research.")
        if len(question) < 10:
            raise Exception("ERROR: Please provide a more detailed question for the analyst to research.")
        if name is None:
            raise Exception("ERROR: Please provide a name for the new analyst.")
        if len(name) < 3:
            raise Exception("ERROR: Please provide a longer name for the new analyst.")
        if len(name) > 50:
            raise Exception("ERROR: Please provide a shorter name for the new analyst.")
        if len(self.analysts) > 0:
            for analyst in self.analysts:
                if analyst.name == name and analyst.goal == question:
                    if analyst.answer is None:
                        new_analyst = analyst
                    else:
                        return ("Response from " + analyst.name + ":\n" + analyst.answer +
                                "\nEvidence provided by " + analyst.name + ":\n" + analyst.evidence)
        if not new_analyst:
            new_analyst = Analyst(self.db, name=name, goal=question)
            self.analysts.append(new_analyst)
        new_analyst.pursue_goal()
        return ("Response from " + name + ":\n" + new_analyst.answer +
                "\nEvidence provided by " + name + ":\n" + new_analyst.evidence)

    def tool_callback(self, response, function_name=None):
        self.messages.append(response)
        self.db.update_last_chat("Processed")
        self.db.add_chat(response["content"], response["role"],
                         "Pending", datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z'),
                         function_name=function_name)

    @staticmethod
    def tool_error_callback(response):
        from pprint import pprint as pp
        pp(response)

    def process_message(self, content, role, status, time, store_message=True):
        if status != "Pending":
            self.db.update_last_chat("Processed")
            return
        if role != "user":
            raise ValueError("Only new user messages can be processed")
        if store_message:
            chat_message = {"content": content, "role": role, "status": status, "time": time}
            self.db.add_chat(chat_message)
        called_tools = True
        while called_tools:
            called_tools = False
            temp_messages = [{"content": self.system_message, "role": "system"}] + self.db.get_database_chat()
            arguments = {"messages": temp_messages, "model": "gpt-4o", "tools": self.tools, "temperature": 0.2}
            chat_response = client.chat.completions.create(**arguments)
            if chat_response.choices[0].message.tool_calls:
                called_tools = True
            if chat_response.choices[0].message.content and not called_tools:
                self.db.update_last_chat("Processed")
                chat_message = {"content": chat_response.choices[0].message.content, "role": "assistant",
                                "status": "Pending", "time": datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')}
                self.db.add_chat(chat_message)
            if called_tools:
                call_new_history = use_tools(chat_response, arguments, function_dict=self.tool_callables, pre_tool_call=True)
                for call in call_new_history:
                    if call["role"] == "assistant":
                        self.db.update_last_chat("Processed")
                        call["status"] = "Pending"
                        call["time"] = datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')
                        if not call["content"]:
                            call["content"] = "Working on that now..."
                        self.db.add_chat(call)
                new_history = use_tools(chat_response, arguments, function_dict=self.tool_callables)
                for call in new_history:
                    if call["role"] != "assistant":
                        self.db.update_last_chat("Processed")
                        call["status"] = "Pending"
                        call["time"] = datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')
                        self.db.add_chat(call)
        self.db.update_last_chat("Processed")

    def finish_tool_calls(self, last_chat):
        new_history = use_tools(last_chat, {"messages": self.db.get_database_chat(), "model": "gpt-4o",
                                            "tools": self.tools, "temperature": 0.2}, function_dict=self.tool_callables)
        for call in new_history:
            if call["role"] != "assistant":
                self.db.update_last_chat("Processed")
                call["status"] = "Pending"
                call["time"] = datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')
                self.db.add_chat(call)
        self.db.update_last_chat("Processed")






import traceback


def run_backend(folder, project_path, storage_path, message_queue, stop_event):
    try:
        from .llm import update_stop_event
        update_stop_event(stop_event)
        from .process_paper import process_paper
        from .database_manager import DatabaseManager
        from .principle_investigator import PrincipalInvestigator
        import subprocess
        import sys
        import time
        start = time.time()
        dm = DatabaseManager(folder, process_paper, project_path, storage_path=storage_path)
        pi = PrincipalInvestigator(dm)
        if time.time() - start > 5*60:
            dm.save_database()
            if sys.platform == "darwin":
                subprocess.Popen(["say", "ScienceAI is ready"])
        while True:
            if message_queue.empty():
                time.sleep(1)
            else:
                message = message_queue.get()
                if message.get("TERMINATE"):
                    print("Terminating backend")
                    break
                elif stop_event.is_set():
                    print("Stop event set. Terminating backend")
                    break
                start = time.time()
                pi.process_message(**message)
                end = time.time()
                if end - start > 10 and sys.platform == "darwin":
                    subprocess.Popen(["say", "New message from ScienceAI"])
                dm.save_database()
    except Exception as e:
        print("Backend error")
        traceback.print_exc()
        raise e




import asyncio
import re
import os
import shutil
import sys
import threading
import time
import urllib
import uuid
import zipfile
import tempfile

from flask_sock import Sock
from flask import Flask, render_template, abort, after_this_request
from .database_manager import DatabaseManager, get_projects
from .backend import run_backend
from multiprocessing import Queue
import atexit
from datetime import datetime
from .llm import client

app = Flask(__name__,
            template_folder='templates',
            static_folder='static')
sock = Sock(app)


database = None
stop_event = None
message_queue = None
thread = None
original_save = None
db_folder = os.path.join(os.path.expanduser('~'), 'Documents', "ScienceAI")
if not os.path.exists(db_folder):
    os.makedirs(db_folder)
path_to_app = os.path.dirname(os.path.abspath(__file__))
path_to_python = sys.executable
script_to_return_to_menu = "<script>window.location.href = '/menu';</script>"


def close():
    global thread
    global stop_event
    global message_queue
    global database
    global original_save
    if thread:
        message_queue.put({"TERMINATE": True})
        stop_event.set()
        thread.join()
    thread = None
    stop_event = None
    message_queue = None
    database = None
    original_save = None


def sanitize_for_id(value):
    """Replace slashes and spaces with dashes."""
    return re.sub("^-|-$", "",
                  value.replace('/', '-').replace(' ', '-').replace('.', '-').replace('--', '-').lower())


app.jinja_env.filters['sanitize_for_id'] = sanitize_for_id


def load_project(project):
    global stop_event
    global database
    global message_queue
    global thread
    global original_save
    if database:
        close()
        return False
    stop_event = threading.Event()
    message_queue = Queue()
    ingest_folder = os.path.join(db_folder, "scienceai_db", project, project.replace(" ", "_")+"_ingest_folder")
    if not os.path.exists(ingest_folder):
        return False
    thread = threading.Thread(target=run_backend, args=(ingest_folder, project, db_folder, message_queue, stop_event))
    thread.start()
    time.sleep(1)
    database = DatabaseManager(ingest_folder, None, project, storage_path=db_folder, read_only_mode=True)
    original_save = database.get_last_save()
    return True


@app.route('/', methods=['GET', 'POST'])
@app.route('/menu', methods=['GET', 'POST'])
def menu():
    from flask import redirect, request
    projects = get_projects(db_folder)
    if request.method == 'POST':
        if "project" in request.form:
            project = request.form["project"]
            if project in projects:
                close()
                result = load_project(project)
                if result:
                    return redirect('/app')
                return redirect('/menu?error=Folder%20not%20found')
            else:
                return redirect('/create?project='+project)
    if request.args.get("error"):
        error = request.args.get("error")
        error = urllib.parse.unquote(error)
        return render_template('menu.html', projects=projects, error=error)
    return render_template('menu.html', projects=projects)


@app.route('/create', methods=['GET', 'POST'])
def create_project():
    from flask import redirect, request
    if request.method == 'GET':
        if "project" in request.args:
            project = request.args["project"]
            return render_template('create.html', project=project)
    if request.method == 'POST':
        if "project" in request.form:
            project = request.form["project"]
            # download the files if they exist in the form under files
            files = request.files.getlist("files")
            # write the files to the ingest
            full_db_folder = os.path.join(db_folder, "scienceai_db")
            os.makedirs(full_db_folder, exist_ok=True)
            project_folder = os.path.join(full_db_folder, project)
            os.makedirs(project_folder, exist_ok=True)
            ingest_folder = os.path.join(project_folder, project.replace(" ", "_")+"_ingest_folder")
            os.makedirs(ingest_folder, exist_ok=True)
            atleast_one_file = False
            for file in files:
                if file.filename == "":
                    continue
                atleast_one_file = True
                file.save(os.path.join(ingest_folder, str(uuid.uuid4())+".pdf"))
            # unzip the files in the zip form if they exist
            zips = request.files.getlist("zips")
            for file in zips:
                if file.filename == "":
                    continue
                zip_name = str(uuid.uuid4())+".zip"
                os.makedirs(os.path.join(ingest_folder, "zip"), exist_ok=True)
                file.save(os.path.join(ingest_folder, "zip", zip_name))
                # using python to unzip the files
                with zipfile.ZipFile(os.path.join(ingest_folder, "zip", zip_name), 'r') as zip_ref:
                    zip_ref.extractall(os.path.join(ingest_folder, "zip"))
                # then delete any non-pdfs or subfolders
                for root, dirs, files in os.walk(os.path.join(ingest_folder, "zip")):
                    for file in files:
                        if file.endswith(".pdf"):
                            if len(dirs) == 0:
                                os.rename(os.path.join(root, file), os.path.join(ingest_folder, file))
                            else:
                                os.rename(os.path.join(root, os.path.join(*dirs), file), os.path.join(ingest_folder, file))
                            atleast_one_file = True
                shutil.rmtree(os.path.join(ingest_folder, "zip"))
            if not atleast_one_file:
                return redirect('/menu?error=No%20files%20uploaded')
            result = load_project(project)
            if result:
                return redirect('/app')
            return redirect('/menu?error=Failed%20to%20create%20project')
    return redirect('/menu?error=Failed%20to%20Create%20Project')


@app.route('/app')
def app_endpoint():
    from flask import redirect
    if database:
        return render_template('app.html')
    return redirect('/menu?error=No%20project%20loaded')


@app.route('/start-database')
def db():
    if not database:
        return script_to_return_to_menu
    db_snippet = database.get_analyst_data_visual("/")
    html_snippet = render_template('db_element.html', data_dict=db_snippet, basepath="Analysts")
    return render_template('db.html', html_snippet=html_snippet)


@app.route('/Analysts', defaults={'path': '/Analysts'})
@app.route('/Analysts/<path:path>')
def update_data(path):
    from urllib.parse import unquote
    if not database:
        return script_to_return_to_menu
    path = unquote(path)
    if path == "/Analysts":
        data_to_return = database.get_analyst_data_visual("/")
    else:
        path = "/Analysts/" + path
        data_to_return = database.get_analyst_data_visual(path)
    if data_to_return:
        return render_template('db_element.html', data_dict=data_to_return, basepath=path)
    return abort(404, description="Resource not found")


@app.route('/download/<path:filepath>')
def download(filepath):
    from flask import send_from_directory, request
    if filepath[0] != '/':
        filepath = "/"+filepath
    target = os.path.join(path_to_app, "io", os.path.basename(filepath))
    shutil.copyfile(filepath, target)

    @after_this_request
    def remove_file(response):
        os.remove(target)
        return response
    dir_path = os.path.dirname(target)
    path = os.path.basename(filepath)
    if request.args.get("attached"):
        return send_from_directory(directory=dir_path, path=path, as_attachment=True)
    else:
        return send_from_directory(directory=dir_path, path=path, as_attachment=False)


@sock.route('/discussion')
def discussion(ws):
    if not database:
        return script_to_return_to_menu
    messages = database.get_database_chat()
    if len(messages) == 0:
        current = str(uuid.uuid4())
    else:
        current = str(hash(str(database.get_database_chat())))
        ws.send(render_template('chat.html', messages=messages))
    while True:
        asyncio.run(database.await_update(timeout=60))
        messages = database.get_database_chat()
        new = str(hash(str(database.get_database_chat())))
        if new != current:
            current = new
            ws.send(render_template('chat.html', messages=messages))


@app.route('/send_message', methods=['POST'])
def send_message():
    from flask import request, render_template
    message = request.form['text']
    new = {"content": message, "time": datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z'), "role": "user",
                       "status": "Pending"}
    message_queue.put(new)
    return render_template('chat_update.html')


@sock.route('/papers')
def papers(ws):
    if not database:
        return script_to_return_to_menu
    papers_dict = database.get_database_papers()
    if len(papers_dict) == 0:
        current = str(uuid.uuid4())
    else:
        current = str(hash(str(database.get_database_papers())))
        ws.send(render_template('papers.html', papers=papers_dict))
    while True:
        asyncio.run(database.await_update(timeout=60))
        new = str(hash(str(database.get_database_papers())))
        if new != current:
            papers_dict = database.get_database_papers()
            current = new
            ws.send(render_template('papers.html', papers=papers_dict))


@app.route('/close_project')
def close_project():
    from flask import redirect, render_template, request
    if request.args.get("confirm"):
        global database
        global message_queue
        if database:
            close()
        database = None
        message_queue = None
        return redirect('/menu')
    last_save = None
    if database:
        last_save = database.get_last_save()
        if not last_save:
            ready = False
        else:
            update_time = database.get_update_time().replace(" ", "_").replace(":", "_")
            ready = last_save.find(update_time) > -1 or last_save == original_save
        messages = database.get_database_chat()
        option = False
        if len(messages) > 0:
            if not ready and messages[-1]["status"] == "Processed":
                option = True
            elif messages[-1]["status"] != "Processed":
                ready = False
        if last_save:
            save_time = datetime.strptime(last_save[-19:], "%Y-%m-%d_%H_%M_%S")
            pretty_time = save_time.strftime('%B %d, %Y %I:%M:%S %p %Z')
        else:
            pretty_time = None
        return render_template('close.html', last_save=pretty_time, ready=ready, option=option)


@app.route('/export_papers')
def export_papers():
    from flask import send_from_directory, request
    from urllib.parse import unquote, quote
    if not database:
        return script_to_return_to_menu
    analystName = request.args.get("analyst", "")
    listName = request.args.get("list", "")
    if len(analystName)+len(listName) == 0:
        listName = None
        analystName = None
    try:
        papers = database.get_all_papers(analyst=analystName, named_list=listName)
    except ValueError:
        return abort(404, description="Resource not found")
    temp_dir = tempfile.mktemp()
    temp_path = os.path.join(temp_dir, "scienceai_paper_export_"+datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
    os.makedirs(temp_path, exist_ok=True)
    selected_fields = unquote(request.args.get("fields")).split(",")
    sep = unquote(request.args.get("seperator", "_"))
    user_defined_tag = unquote(request.args.get("userDefinedTag", ""))
    inj = "INJECT-TITLE"
    names = []
    for paper in papers:
        title = "NA"
        data = database.get_paper_data(paper.get("paper_id"))
        metadata = data.get("metadata", {})
        name = ""
        for field in selected_fields:
            if field == 'User Defined Tag':
                name += user_defined_tag + sep
            elif field == 'ScienceAI List':
                if listName:
                    name += listName + sep
                else:
                    name += "NA" + sep
            elif field == 'DOI':
                if "DOI" in metadata:
                    name += metadata["DOI"] + sep
                else:
                    name += "NA" + sep
            elif field == 'Date of Publication':
                if "created" in metadata:
                    name += metadata["created"]["date-time"][:10] + sep
                else:
                    name += "NA" + sep
            elif field == 'First Author':
                if "authors" in metadata:
                    name += metadata["authors"][0]["given"] + "-" + metadata["authors"][0]["family"] + sep
                else:
                    name += "NA" + sep
            elif field == 'Title':
                if "title" in metadata:
                    name += inj + sep
                    title = metadata["title"][0]
                else:
                    name += "NA" + sep
            elif field == 'Journal':
                if "container-title" in metadata:
                    name += metadata["container-title"][0] + sep
                else:
                    name += "NA" + sep
        name = name[:-1]
        revert = name
        replacements = {
            '<': '_lt_',
            '>': '_gt_',
            ':': '_colon_',
            '"': '_quote_',
            "'": '_quote_',
            '/': '_slash_',
            '\\': '_backslash_',
            '|': '_pipe_',
            '?': '_question_',
            '*': '_asterisk_',
            '.': '_period_',
        }

        # Replace invalid characters
        for invalid_char, replacement in replacements.items():
            name = name.replace(invalid_char, replacement)
            title = title.replace(invalid_char, replacement)

        # Replace any remaining invalid characters with an underscore
        name = re.sub(r'[^\w\-_\. ]', '', name)
        title = re.sub(r'[^\w\-_\. ]', '', title)

        name = name.replace(inj, title)+".pdf"
        if len(name) > 255:
            # chop off the end of the title
            short_title = title[:255-len(name)]
            name = revert.replace(inj, short_title)+".pdf"
        if len(name) > 255:
            # chop off the end of the title
            short_title = title[:255-len(name)]
            name = revert.replace(inj, short_title)[:251]+".pdf"

        if name in names:
            name = name[:251]+"_"+str(names.count(name)+1)+".pdf"

        names.append(name)

        shutil.copyfile(database.get_paper_pdf(paper.get("paper_id")), os.path.join(temp_path, name))
    source = temp_path
    zip_name = "scienceai_paper_export_"+datetime.now().strftime('%Y-%m-%d_%H-%M-%S')+".zip"
    destination = os.path.join(path_to_app, "io", zip_name)
    base = os.path.basename(destination)
    name = base.split('.')[0]
    format = base.split('.')[1]
    archive_from = os.path.dirname(source)
    archive_to = os.path.basename(source.strip(os.sep))
    shutil.make_archive(name, format, archive_from, archive_to)
    shutil.move('%s.%s' % (name, format), destination)
    shutil.rmtree(temp_dir)

    dir_path = os.path.dirname(destination)
    path = os.path.basename(destination)

    @after_this_request
    def remove_file(response):
        os.remove(destination)
        return response
    return send_from_directory(directory=dir_path, path=path, as_attachment=True)


@app.route('/save')
def save():
    from flask import redirect
    if not database:
        return script_to_return_to_menu
    database.save_database()
    return redirect('/app')


@app.route('/save_project')
def save_project():
    from flask import render_template
    if not database:
        return script_to_return_to_menu
    last_save = database.get_last_save()
    if last_save:
        update_time = database.get_update_time().replace(" ", "_").replace(":", "_")
        ready = last_save.find(update_time) > -1 or last_save == original_save
    else:
        ready = False
    messages = database.get_database_chat()
    option = False
    if len(messages) > 0:
        if not ready and messages[-1]["status"] == "Processed":
            option = True
    if last_save:
        save_time = datetime.strptime(last_save[-19:], "%Y-%m-%d_%H_%M_%S")
        pretty_time = save_time.strftime('%B %d, %Y %I:%M:%S %p %Z')
    else:
        pretty_time = None
    return render_template('save.html', last_save=pretty_time, option=option)


@app.route('/download_save')
def download_save():
    from flask import send_from_directory
    if not database:
        return script_to_return_to_menu
    save_path = database.get_last_save(path=True)
    temp_dir = tempfile.mktemp()
    project = os.path.basename(database.project_path)
    source = os.path.join(temp_dir, project)
    shutil.copytree(save_path, source)
    zip_name = project.replace(" ", "_")+"_scienceai_save_"+datetime.now().strftime('%Y-%m-%d_%H-%M-%S')+".zip"
    destination = os.path.join(path_to_app, "io", zip_name)
    base = os.path.basename(destination)
    name = base.split('.')[0]
    format = base.split('.')[1]
    archive_from = os.path.dirname(source)
    archive_to = os.path.basename(source.strip(os.sep))
    shutil.make_archive(name, format, archive_from, archive_to)
    shutil.move('%s.%s' % (name, format), destination)

    dir_path = os.path.dirname(destination)
    path = os.path.basename(destination)

    @after_this_request
    def remove_file(response):
        os.remove(destination)
        return response
    return send_from_directory(directory=dir_path, path=path, as_attachment=True)


@app.route('/download_analysis')
def download_analysis():
    from flask import send_from_directory
    if not database:
        return script_to_return_to_menu
    analysis_path = database.combine_analyst_tool_trackers()
    project = os.path.basename(database.project_path)
    destination = os.path.join(path_to_app, "io", project.replace(" ", "_")+"_scienceai_analysis_"+datetime.now().strftime('%Y-%m-%d_%H-%M-%S')+".csv")
    os.rename(analysis_path, destination)
    dir_path = os.path.dirname(destination)
    path = os.path.basename(destination)

    @after_this_request
    def remove_file(response):
        os.remove(destination)
        return response
    return send_from_directory(directory=dir_path, path=path, as_attachment=True)


@app.route('/load_checkpoint', methods=['POST'])
def load_save():
    from flask import request, redirect
    if database:
        close_project()
        return script_to_return_to_menu
    save_file = request.files["checkpoint"]
    temp_dir = tempfile.mktemp()
    os.makedirs(temp_dir, exist_ok=True)
    project = request.form["project"]
    save_path = os.path.join(temp_dir, "save.zip")
    unzip_folder_path = os.path.join(temp_dir, project)
    folder_path = os.path.join(unzip_folder_path, project)
    save_file.save(save_path)
    shutil.unpack_archive(save_path, unzip_folder_path)
    found_project_name = None
    for dir in os.listdir(temp_dir):
        found_project_name = os.path.basename(dir)
    unzip_folder_path = os.path.join(unzip_folder_path, found_project_name)
    os.rename(unzip_folder_path, folder_path)
    projects_folder = os.path.join(db_folder, "scienceai_db")
    if not os.path.exists(projects_folder):
        os.makedirs(projects_folder)
    project_path = os.path.join(projects_folder, project)
    if os.path.exists(project_path):
        if request.form.get("overwrite"):
            shutil.rmtree(project_path)
        else:
            return redirect('/menu?error=Project%20already%20exists')
    shutil.move(folder_path, project_path)
    shutil.rmtree(temp_dir)
    result = load_project(project)
    if result:
        return redirect('/app')
    return redirect('/menu?error=Failed%20to%20load%20project')


@app.route('/delete_project', methods=['POST'])
def delete_project():
    from flask import redirect, request
    if database:
        return redirect('/app')
    project = request.form["project"]
    project_path = os.path.join(db_folder, "scienceai_db")
    checkpoints = []
    for dir in os.listdir(project_path):
        if dir.find(project+"_-checkpoint-_") > -1:
            checkpoints.append(os.path.join(project_path, dir))
    for checkpoint in checkpoints:
        shutil.rmtree(checkpoint)
    shutil.rmtree(os.path.join(project_path, project))
    return redirect('/menu')


atexit.register(close)


def main():
    import logging
    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)
    # print a clickable link to the user to open the app by navigating to the link
    print("ScienceAI is running. Please open the following link in your browser to access the application:")
    print("\033]8;;http://localhost:4242\ahttp://localhost:4242\033]8;;\a")
    app.run(host='localhost', port=4242, debug=False)


if __name__ == '__main__':
    main()



import time
import sys
import subprocess
from .process_paper import process_paper
from .database_manager import DatabaseManager
from .principle_investigator import PrincipalInvestigator
from datetime import datetime


if __name__ == "__main__":
    folder = sys.argv[1]
    project_path = sys.argv[2]
    storage_path = sys.argv[3]
    dm = DatabaseManager(folder, process_paper, project_path, storage_path=storage_path)
    print("Assistant: Hello, I am ScienceAI. I first need to make sure all your papers are loaded into the "
          "system before I can help you. I will let you know when I am ready to answer your questions. ")
    messages = dm.get_database_chat()
    offset = len(messages)
    if offset == 0:
        print("Assistant: All papers have been loaded into the system. How can I help you today?")
    pi = PrincipalInvestigator(dm)
    for message in messages:
        if message["role"] == "user":
            print("User: " + message["content"])
        else:
            print("Assistant: " + message["content"])
    while True:
        messages = dm.get_database_chat()
        for message in messages[offset:]:
            if message["role"] == "user":
                print("User: " + message["content"])
            else:
                print("Assistant: " + message["content"])
        input_message = input("User: ")
        if input_message == "exit":
            break
        offset += 1
        input_dict = {"content": input_message, "role": "user", "status": "Pending",
                      "time": datetime.now().strftime('%B %d, %Y %I:%M:%S %p %Z')}
        start = time.time()
        pi.process_message(**input_dict)
        end = time.time()
        if end - start > 10:
            subprocess.Popen(["say", "New message from ScienceAI"])




from .llm import client, use_tools
from .database_manager import DatabaseManager
from .data_extractor import generate_schema, extract_data, schema_to_tool
import os

short_id = {}

path_to_app = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(path_to_app, "analyst_base_prompt.txt"), "r") as f:
    analyst_system = f.read()


def reflect_on_evidence(goal, answer, evidence, retries=3):
    system_message = ("The analyst has answered the following question / goal with evidence. "
                      "You are a thoughtful Researcher, evaluate the evidence and "
                      "determine if the goal has been achieved or the question has been answered.")
    user_message = f"My goal/question: {goal}\n\nMy answer is:\n{answer}\n\nMy evidence:\n{evidence}."

    messages = [
        {
            "role": "system",
            "content": system_message
        },
        {
            "role": "user",
            "content": user_message
        }
    ]

    arguments = {"messages": messages, "model": "gpt-4o", }

    chat_response = client.chat.completions.create(**arguments)

    thoughts = chat_response.choices[0].message.content

    messages.append({"role": "assistant", "content": thoughts})

    tools = [
        {
            "type": "function",
            "function": {
                "name": "check_completed_goal",
                "description": "Checks if the goal has been completed or the question has "
                               "been answered and the evidence is sufficient.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "resolved": {
                            "type": "boolean",
                            "description": "Whether the goal has been completed or the question has been answered."
                        }
                    },
                },
                "required": ["resolved"]
            }
        }
    ]

    arguments = {"messages": messages, "model": "gpt-4o", "tools": tools,
                 "tool_choice": {"type": "function", "function": {"name": "check_completed_goal"}}}

    retry = 0
    valid_calls = []
    while valid_calls == [] and retry < retries:
        if retry > 0:
            print("Retrying...")
        chat_response = client.chat.completions.create(**arguments)
        if chat_response.choices[0].message.tool_calls:
            valid_calls = use_tools(chat_response, arguments, call_functions=False)
            if valid_calls:
                for call in valid_calls:
                    if call["name"] == "check_completed_goal":
                        if not call["parameters"]["resolved"]:
                            return thoughts
                        else:
                            return ""
        retry += 1
    return thoughts


# Analyst Module
class Analyst:
    def __init__(self, db: DatabaseManager, analyst_dict={}, name="", goal="", attempts=5):
        self.db = db
        self.attempts = attempts
        if analyst_dict and name and goal:
            raise ValueError("Can not provide both analyst_dict and name and goal.")
        if "name" in analyst_dict and "goal" in analyst_dict:
            self.name = analyst_dict["name"]
            self.goal = analyst_dict["goal"]
            self.answer = analyst_dict.get("answer", None)
            self.evidence = analyst_dict.get("evidence", None)
        if name and goal:
            self.name = name
            self.goal = goal
            self.answer = None
            self.evidence = None
        try:
            self.db.get_analyst_metadata(self.name)
        except ValueError:
            self.db.create_analyst(name, goal, other={"goal_achieved": False})
        self.get_all_papers()
        self.tool_callables = {
            "get_all_papers": self.get_all_papers,
            "create_named_paper_list": self.create_named_paper_list,
            "get_named_paper_list": self.get_named_paper_list,
            "create_data_collection_request": self.create_data_collection_request,
            "complete_goal_by_answering_question_with_evidence": self.complete_goal_by_answering_question_with_evidence
        }
        self.tools = [
            self.get_all_papers(return_tool=True),
            self.create_named_paper_list(None, None, return_tool=True),
            self.get_named_paper_list(None, return_tool=True),
            self.create_data_collection_request(None, None, return_tool=True),
            self.complete_goal_by_answering_question_with_evidence(None, None, return_tool=True)
        ]
        self.follow_up_answer = None
        self.follow_up_evidence = None
        messages = self.db.get_analyst_context(self.name)
        answer_attempts = [message for message in messages if message["role"] == "tool" and message["name"] == "complete_goal_by_answering_question_with_evidence"]
        self.answer_attempts = len(answer_attempts)

    def get_context(self):
        return self.db.get_analyst_context(self.name)

    def get_all_papers(self, all=True, return_tool=False):
        if return_tool:
            return {
                "type": "function",
                "function": {
                    "name": "get_all_papers",
                    "description": "Prints all papers in the database.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "all": {
                                "type": "boolean",
                                "description": "Whether to return all papers."
                            }
                        },
                    },
                    "required": ["all"],
                }
            }
        output = {}
        papers = self.db.get_all_papers_data()
        for paper in papers:
            short_id[paper['database']['paper_id'][:10]] = paper['database']['paper_id']
            output[paper['database']['paper_id'][:10]] = paper['metadata']['title'][0]
        return output

    def create_named_paper_list(self, name="", paper_ids=[], return_tool=False):
        if return_tool:
            return {
                "type": "function",
                "function": {
                    "name": "create_named_paper_list",
                    "description": "Creates a permanent list of papers (this can not me mutate later).",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string",
                                "description": "The name of the list."
                            },
                            "paper_ids": {
                                "type": "array",
                                "description": "The IDs of the papers to add to the list.",
                                "items": {
                                    "type": "string",
                                    "description": "The ID of the paper."
                                }
                            }
                        },
                    },
                    "required": ["name", "paper_ids"],
                }
            }
        if self.db.get_all_papers(analyst=self.name, named_list=name):
            raise ValueError("List '" + name + "' already exists.")
        for paper_id in paper_ids:
            self.db.add_paper_to_list(short_id[paper_id], self.name, name)
        return "List named '" + name + "' created with papers: " + str(paper_ids)

    def get_named_paper_list(self, name="", return_tool=False):
        if return_tool:
            return {
                "type": "function",
                "function": {
                    "name": "get_named_paper_list",
                    "description": "Gets the papers in a list.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string",
                                "description": "The name of the list."
                            }
                        },
                    },
                    "required": ["name"],
                }
            }
        papers = self.db.get_all_papers_data(analyst=self.name, named_list=name)
        output = {}
        for paper in papers:
            short_id[paper['paper_id'][:10]] = paper['database']['paper_id']
            output[paper['paper_id'][:10]] = paper['metadata']['title'][0]
        return output

    def create_data_collection_request(self, collection_name="", collection_goal="",
                                       target_list=None, return_tool=False):
        if return_tool:
            return {
                "type": "function",
                "function": {
                    "name": "create_data_collection_request",
                    "description": "Creates a schema for a data collection request.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "collection_name": {
                                "type": "string",
                                "description": "The name of the data collection request."
                            },
                            "collection_goal": {
                                "type": "string",
                                "description": "The goal of the data collection request."
                            },
                            "target_list": {
                                "type": "string",
                                "description": "The name of the list of papers to collect data from."
                            }
                        },
                    },
                    "required": ["collection_name", "collection_goal"],
                }
            }

        if target_list:
            try:
                papers = self.db.get_all_papers_data(analyst=self.name, named_list=target_list)
            except ValueError:
                raise ValueError("List not found.")
        else:
            papers = self.db.get_all_papers_data()

        summaries = ""
        for paper in papers:
            summaries += paper["metadata"]["title"][0] + "\n\nSummary: " + paper["summary"] + "\n\n\n"
        schema = generate_schema(summaries, goal=collection_name+" - "+collection_goal)
        tool = schema_to_tool(schema)

        results = {}
        tracker = self.db.add_analyst_tool_tracker(self.name, collection_name)
        for paper in papers:
            result = extract_data(tool, paper["cleaned_text"])
            self.db.update_analyst_tool_tracker(tracker, paper["database"]["paper_id"], result)
            short_id[paper["database"]["paper_id"][:10]] = paper["database"]["paper_id"]
            results[paper["database"]["paper_id"][:10]] = result
        self.db.convert_analyst_tool_tracker(self.name, collection_name)
        return results

    def complete_goal_by_answering_question_with_evidence(self, answer="", evidence="", return_tool=False):
        if return_tool:
            return {
                "type": "function",
                "function": {
                    "name": "complete_goal_by_answering_question_with_evidence",
                    "description": "Completes the analyst's goal by answering a question with evidence.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "answer": {
                                "type": "string",
                                "description": "The answer to the question."
                            },
                            "evidence": {
                                "type": "string",
                                "description": "The evidence you used to answer the question."
                            }
                        },
                    },
                    "required": ["answer", "evidence"],
                }
            }
        thoughts = reflect_on_evidence(self.goal, answer, evidence)
        self.answer_attempts += 1
        if thoughts == "":
            self.answer = answer
            self.evidence = evidence
            return "Goal achieved:\n" + answer + "\n\nEvidence:\n" + evidence
        return "Goal not achieved. Here are some thoughts on why: " + thoughts

    def answer_followup_question(self, answer="", evidence="", return_tool=False):
        if return_tool:
            return {
                "type": "function",
                "function": {
                    "name": "answer_followup_question",
                    "description": "Answers a follow-up question with evidence.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "answer": {
                                "type": "string",
                                "description": "The answer to the question."
                            },
                            "evidence": {
                                "type": "string",
                                "description": "The evidence to answer the question."
                            }
                        },
                    },
                    "required": ["answer", "evidence"],
                }
            }
        thoughts = reflect_on_evidence(self.goal, answer, evidence)
        if thoughts == "":
            self.follow_up_answer = answer
            self.follow_up_evidence = evidence
            return "Question answered:\n" + answer + "\n\nEvidence:\n" + evidence
        return "Question not answered. Here are some thoughts on why: " + thoughts

    def pursue_goal(self):
        messages = self.db.get_analyst_context(self.name)
        if not messages:
            user_message = f"Here is your goal/question: {self.goal}\n\n"
            messages = [
                {
                    "role": "system",
                    "content": analyst_system
                },
                {
                    "role": "user",
                    "content": user_message
                }
            ]
            for message in messages:
                self.db.add_analyst_context(self.name, message)
        while not self.answer:
            messages = self.db.get_analyst_context(self.name)
            arguments = {"messages": messages, "model": "gpt-4o", "tools": self.tools, "temperature": 0.2}
            chat_response = client.chat.completions.create(**arguments)
            new_history = use_tools(chat_response, arguments, function_dict=self.tool_callables)
            for call in new_history:
                self.db.add_analyst_context(self.name, call)
                messages.append(call)
            if self.answer_attempts > self.attempts and not self.answer:
                self.answer = ("The analyst has not been able to answer the question in the allotted attempts. "
                               "Refine the goal and make sure it is specific and longer to help the next analyst "
                               "succeed where this one failed. You should remind it that when it creates its "
                               "data collection requests it should include details on how to avoid those pitfalls.")
                reasons = [message["content"] for message in messages if message["role"] == "tool" and message["name"] == "complete_goal_by_answering_question_with_evidence"]
                self.evidence = ("Here are the reasons the analyst failed to reach its goal "
                                 "after ") + str(self.attempts) + " attempts:" + "\n\n" + "\n\n".join(reasons)
        self.db.add_analyst_metadata(self.name,
                                     {"goal_achieved": True, "answer": self.answer, "evidence": self.evidence})



import os
import shutil
import time
from datetime import datetime
import asyncio
import hashlib
import pandas as pd
from pandas import json_normalize
import dictdatabase as DDB

lock_files = []


def sha256sum(filename):
    h = hashlib.sha256()
    b = bytearray(128*1024)
    mv = memoryview(b)
    with open(filename, 'rb', buffering=0) as f:
        while n := f.readinto(mv):
            h.update(mv[:n])
    return h.hexdigest()


def get_projects(storage_path):
    projects = []
    if os.path.basename(storage_path) != "scienceai_db":
        storage_path = os.path.join(storage_path, "scienceai_db")
    if not os.path.exists(storage_path):
        return projects
    for project in os.listdir(storage_path):
        if os.path.isdir(os.path.join(storage_path, project)) and project.find("_-checkpoint-_") == -1:
            projects.append(project)
    return projects


# Paper Manager
class DatabaseManager:
    def __init__(self, input_pdf_directory, processor, project_name, storage_path=None, auto_prune=False,
                 read_only_mode=False, lock_timeout=60):
        if read_only_mode:
            self.auto_prune = False
        else:
            self.auto_prune = auto_prune
        self.lock_timeout = lock_timeout
        self.read_only_mode = read_only_mode
        self.processor = processor
        self.input_pdf_directory = input_pdf_directory
        if storage_path:
            if not os.path.exists(storage_path):
                os.makedirs(storage_path)
            self.storage_path = os.path.join(storage_path, "scienceai_db")
            if not os.path.exists(self.storage_path):
                os.makedirs(self.storage_path)
        else:
            self.storage_path = os.path.join(self.input_pdf_directory, "scienceai_db")
            if not os.path.exists(self.storage_path):
                os.makedirs(self.storage_path)
        self.project_path = os.path.join(self.storage_path, project_name)
        if not os.path.exists(self.project_path):
            os.makedirs(self.project_path)
        self.papers_pdf_path = os.path.join(self.project_path, "papers_pdf")
        if not os.path.exists(self.papers_pdf_path):
            os.makedirs(self.papers_pdf_path)
        self.db_path = os.path.join(self.project_path, "scienceai_ddb")
        self.db_path_update = os.path.join(self.project_path, project_name + "_update_time.txt")
        self.update_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        with open(self.db_path_update, "w") as f:
            f.write(self.update_time)
        self.default_schema = ["metadata", "papers", "pi_context"]
        self.project_name = project_name
        if not read_only_mode:
            self.__initialize_db__()
    
    def log_update(func):
        def wrapper(self, *args, **kwargs):
            if self.read_only_mode:
                raise ValueError("Database is in read only mode")
            self.update_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            with open(self.db_path_update, "w") as f:
                f.write(self.update_time)
            result = func(self, *args, **kwargs)
            return result
        return wrapper

    async def await_update(self, timeout=None):
        start = time.time()
        while timeout is None or time.time() - start < timeout:
            await asyncio.sleep(1)
            with open(self.db_path_update, "r") as f:
                update_time = f.read()
            if len(update_time) > 11:
                if update_time != self.update_time:
                    self.update_time = update_time
                    break

    def update_update_time(self):
        with open(self.db_path_update, "r") as f:
            temp_update_time = f.read()
        if len(temp_update_time) > 11:
            self.update_time = temp_update_time

    def get_update_time(self):
        self.update_update_time()
        return self.update_time

    @log_update
    def __initialize_db__(self):
        DDB.config.storage_directory = self.db_path
        if not DDB.at("metadata").exists():
            DDB.at("metadata").create({})
        with DDB.at("metadata").session() as (session, metadata):
            metadata[self.project_name] = {"loaded": True}
            for project in os.listdir(self.storage_path):
                if project != self.project_name:
                    if os.path.isdir(os.path.join(self.storage_path, project)):
                        if project not in metadata:
                            metadata[project] = {"loaded": False}
                        else:
                            metadata[project]["loaded"] = False
            session.write()

    @log_update
    def remove_old_default_messages(self, default_messages):
        if not DDB.at("chat").exists():
            DDB.at("chat").create({})
        if default_messages and DDB.at("chat", key="messages").exists():
            current_messages = DDB.at("chat", key="messages").read()
            index = len(current_messages) - len(default_messages)
            last_message = current_messages[index:]
            last_content = [message["content"] for message in last_message]
            if last_content == default_messages:
                with DDB.at("chat", key="messages").session() as (session, messages):
                    messages = current_messages[:index]
                    session.write()

    @log_update
    def ingest_paper(self, pdf_path):
        if not os.path.exists(pdf_path):
            raise ValueError(f"File {pdf_path} does not exist")
        paper_id = sha256sum(pdf_path)
        filename = paper_id + ".pdf"
        stored_pdf_path = os.path.join(self.papers_pdf_path, filename)
        if not os.path.exists(stored_pdf_path):
            shutil.copy(pdf_path, stored_pdf_path)
        if not DDB.at('papers').exists():
            DDB.at('papers').create({})
        with DDB.at('papers').session() as (session, papers):
            if paper_id not in papers:
                papers[paper_id] = {"pdf_path": stored_pdf_path, "paper_id": paper_id}
            else:
                papers[paper_id].update({"pdf_path": stored_pdf_path})
            session.write()
        return paper_id

    def ingest_papers(self):
        found_papers = []
        for file in os.listdir(self.input_pdf_directory):
            if file.endswith(".pdf"):
                pdf_path = os.path.join(self.input_pdf_directory, file)
                found_papers.append(self.ingest_paper(pdf_path))
        if self.auto_prune and DDB.at("papers").exists():
            papers = DDB.at("papers").read()
            prune_papers = [paper['paper_id'] for paper in papers if paper['paper_id'] not in found_papers]
            for paper_id in prune_papers:
                self.prune_paper(paper_id)
        return found_papers

    @log_update
    def prune_paper(self, paper_id):
        if DDB.at("papers", key=paper_id).exists():
            with DDB.at("papers").session() as (session, papers):
                paths = papers[paper_id]
                if 'pdf_path' in paths and os.path.exists(paths['pdf_path']):
                    os.remove(paths['pdf_path'])
                if 'json_path' in paths and os.path.exists(paths['json_path']):
                    os.remove(paths['json_path'])
                del papers[paper_id]
                session.write()

    @log_update
    def update_paper(self, paper_id, paper_metadata):
        if DDB.at("papers", key=paper_id).exists():
            with DDB.at("papers").session() as (session, papers):
                papers[paper_id].update(paper_metadata)
                session.write()

    def get_paper(self, paper_id):
        if not DDB.at("papers", key=paper_id).exists():
            raise ValueError(f"Paper with id {paper_id} not found")
        return DDB.at("papers", key=paper_id).read()

    def get_paper_pdf(self, paper_id):
        paper = self.get_paper(paper_id)
        return paper.get("pdf_path", None)

    def get_paper_json(self, paper_id):
        paper = self.get_paper(paper_id)
        return paper.get("json_path", None)

    @log_update
    def store_paper_json(self, paper_id, dict_data):
        paper = self.get_paper(paper_id)
        if not DDB.at(paper_id).exists():
            DDB.at(paper_id).create(dict_data)
        with DDB.at("papers").session() as (session, papers):
            papers[paper_id] = paper
            if dict_data.get("metadata"):
                added = {}
                if dict_data["metadata"].get("title"):
                    added["Title"] = dict_data["metadata"]["title"][0]
                if dict_data["metadata"].get("created") and dict_data["metadata"]["created"].get("date-time"):
                    added["Date"] = dict_data["metadata"]["created"]["date-time"][:10]
                if dict_data["metadata"].get("author"):
                    added["Authors"] = dict_data["metadata"]["author"][0]["given"] + " " + dict_data["metadata"]["author"][0][
                        "family"]
                papers[paper_id].update(added)
            session.write()
        return True

    @log_update
    def process_paper(self, paper_id):
        """ Processes the paper

        Returns:
            string: path to the paper

        """
        pdf_path = self.get_paper_pdf(paper_id)
        if not DDB.at(paper_id).exists():
            processed_paper = self.processor(pdf_path)
            self.store_paper_json(paper_id, processed_paper)

    def process_all_papers(self):
        """ Processes all the papers

        Returns:
            string: path to the paper

        """
        print("Processing all papers")
        paper_ids = list(DDB.at("papers").read().keys())
        for paper_id in paper_ids:
            print(f"Processing paper {paper_id}")
            self.process_paper(paper_id)
        return True

    @log_update
    def add_chat(self, message):
        with DDB.at("chat").session() as (session, chat):
            if "messages" not in chat:
                chat["messages"] = []
            chat["messages"].append(message)
            session.write()
        return True

    @log_update
    def update_last_chat(self, status, full_update=None):
        with DDB.at("chat").session() as (session, chat):
            if "messages" in chat:
                if full_update:
                    chat["messages"][-1] = full_update
                chat["messages"][-1]["status"] = status
                session.write()
        return True

    @log_update
    def create_analyst(self, name, goal, other={}):
        if not DDB.at("Analysts").exists():
            DDB.at("Analysts").create({})
        if not DDB.at(name).exists():
            DDB.at(name).create({})
        with DDB.at("Analysts").session() as (session, analysts):
            if name not in analysts:
                analysts[name] = {"goal": goal, **other}
            session.write()
        with DDB.at(name).session() as (session, analyst):
            analyst["context"] = []
            session.write()
        return True

    @log_update
    def add_analyst_context(self, name, analyst_context):
        if not DDB.at(name).exists():
            DDB.at(name).create({})
        if not DDB.at(name).exists():
            raise ValueError(f"Analyst {name} not found")
        with DDB.at(name).session() as (session, analyst):
            analyst["context"].append(analyst_context)
            session.write()
        return True

    @log_update
    def add_analyst_tool_tracker(self, analyst_name, tool_name, json_data={}):
        tool_time = datetime.now().strftime("%Y-%m-%d_%H_%M_%S")
        tool_fullname = analyst_name+"_/"+tool_name + "_" + tool_time
        if not DDB.at(tool_fullname).exists():
            DDB.at(tool_fullname).create(json_data)
        else:
            with DDB.at(tool_fullname).session() as (session, tool):
                tool.update(json_data)
                session.write()
        with DDB.at("Analysts").session() as (session, analysts):
            if analyst_name not in analysts:
                raise ValueError(f"Analyst {analyst_name} not found")
            if "tools" not in analysts[analyst_name]:
                analysts[analyst_name]["tools"] = []
            analysts[analyst_name]["tools"].append({"tool_name": tool_name, "json_path": tool_fullname + ".json", "hidden": True})
            session.write()
        return tool_fullname

    @log_update
    def convert_analyst_tool_tracker(self, analyst_name, tool_name):
        if not DDB.at("Analysts", key=analyst_name).exists():
            raise ValueError(f"Analyst {analyst_name} not found")
        data = None
        with DDB.at("Analysts", key=analyst_name).session() as (session, analyst):
            for i, tool in enumerate(analyst["tools"]):
                if tool["tool_name"] == tool_name:
                    data_path = tool["json_path"]
                    data = DDB.at(data_path.replace(".json", "")).read()
                    data = list({k: {**{"id": k[10:]}, **v} for k, v in data.items()}.values())
                    flat_data = json_normalize(data)
                    csv_path = data_path.replace(".json", ".csv")
                    csv_folder = os.path.join(self.project_path, "csv_files")
                    if not os.path.exists(csv_folder):
                        os.makedirs(csv_folder)
                    csv_path = os.path.join(csv_folder, os.path.basename(csv_path))
                    flat_data.to_csv(csv_path, index=False)
                    analyst["tools"][i]["csv_path"] = csv_path
            if not data:
                raise ValueError(f"Tool {tool_name} not found for analyst {analyst_name}")
            # dict comprehension to add an id key of the first 10 characters to each value
            session.write()
        return csv_path

    def combine_analyst_tool_trackers(self):
        if not DDB.at("Analysts").exists():
            df = pd.DataFrame({"Notes": ["No Analysts have been created yet. After they are created, results of their"
                                         " data extraction will be combined and accessible under the "
                                         "'Extracted Data' tab."]})
            df.to_csv(os.path.join(self.project_path, "merged_analyst_tools.csv"), index=False)
            return os.path.join(self.project_path, "merged_analyst_tools.csv")
        csv_paths = {}
        for analyst_name in DDB.at("Analysts").read():
            analyst = DDB.at("Analysts", key=analyst_name).read()
            if "tools" in analyst:
                for tool in analyst["tools"]:
                    if "csv_path" in tool:
                        csv_paths[tool["csv_path"]] = {"name": tool["tool_name"], "analyst": analyst_name}
        if not csv_paths:
            df = pd.DataFrame({"Notes": ["No Analysts have extracted data yet. After they do, the results will be "
                                         "combined and accessible under the 'Extracted Data' tab."]})
            df.to_csv(os.path.join(self.project_path, "merged_analyst_tools.csv"), index=False)
            return os.path.join(self.project_path, "merged_analyst_tools.csv")

        columns = []
        double_columns = {}
        triple_columns = []
        for csv in csv_paths.keys():
            df = pd.read_csv(csv)
            for col in df.columns:
                if col == "id":
                    continue
                elif col not in columns:
                    columns.append(col)
                elif col in columns:
                    if col not in double_columns:
                        double_columns[col+"_"+csv_paths[csv]["analyst"]] = col
                    else:
                        triple_columns.append(col)
        double_columns = list(double_columns.values())

        merged_df = pd.DataFrame()

        for csv, meta in csv_paths.items():
            df = pd.read_csv(csv)
            remap = {}
            for col in df.columns:
                if col in double_columns:
                    remap[col] = col + "_" + meta["analyst"]
                if col in triple_columns:
                    remap[col] = col + "_" + meta["analyst"] + "_" + meta["name"]
            df.rename(columns=remap, inplace=True)
            if merged_df.empty:
                merged_df = df
            else:
                merged_df = pd.merge(merged_df, df, how="outer", on="id")

        merged_df.to_csv(os.path.join(self.project_path, "merged_analyst_tools.csv"), index=False)
        return os.path.join(self.project_path, "merged_analyst_tools.csv")

    @log_update
    def update_analyst_tool_tracker(self, json_path, key, update_data, overwrite_list=False):
        json_path = json_path.replace(".json", "")
        if not DDB.at(json_path).exists():
            raise ValueError(f"File {json_path} does not exist")
        with DDB.at(json_path).session() as (session, data):
            if key not in data:
                data[key] = update_data
            elif isinstance(data[key], dict):
                data[key].update(update_data)
            elif isinstance(data[key], list):
                if overwrite_list:
                    data[key] = update_data
                else:
                    if isinstance(update_data, list):
                        data[key].extend(update_data)
                    else:
                        data[key].append(update_data)
            session.write()
        return True

    @log_update
    def add_analyst_metadata(self, name, metadata):
        if not DDB.at("Analysts").exists():
            DDB.at("Analysts").create({})
        if not DDB.at("Analysts", key=name).exists():
            raise ValueError(f"Analyst {name} not found")
        with DDB.at("Analysts").session() as (session, analysts):
            analysts[name].update(metadata)
            session.write()
        return True

    @log_update
    def add_paper_to_list(self, paper_id, analyst, name_of_list):
        if not DDB.at("papers", key=paper_id).exists():
            raise ValueError(f"Paper with id {paper_id} not found")
        if not DDB.at("Analysts").exists():
            DDB.at("Analysts").create({})
        if not DDB.at("Analysts", key=analyst).exists():
            raise ValueError(f"Analyst {analyst} not found")
        with DDB.at("papers").session() as (session, papers):
            if analyst not in papers[paper_id]:
                papers[paper_id][analyst] = [name_of_list]
            else:
                if name_of_list not in papers[paper_id][analyst]:
                    papers[paper_id][analyst].append(name_of_list)
            session.write()
        return True

    @log_update
    def remove_paper_from_list(self, paper_id, analyst, name_of_list=None):
        if not DDB.at("papers", key=paper_id).exists():
            raise ValueError(f"Paper with id {paper_id} not found")
        if not DDB.at("Analysts").exists():
            DDB.at("Analysts").create({})
        with DDB.at("papers").session() as (session, papers):
            if analyst in papers[paper_id]:
                if name_of_list is None:
                    papers[paper_id][analyst] = []
                elif name_of_list in papers[paper_id][analyst]:
                    papers[paper_id][analyst].remove(name_of_list)
            session.write()
        return True

    def get_all_tool_trackers_for_analyst(self, analyst_name):
        analysts = DDB.at("Analysts").read()
        if not analysts:
            return []
        if analyst_name not in analysts:
            raise ValueError(f"Analyst {analyst_name} not found")
        if "tools" not in analysts[analyst_name]:
            return []
        files = {}
        tools = {}
        for tool in analysts[analyst_name]["tools"]:
            csv_path = None
            if "csv_path" in tool:
                csv_path = tool["csv_path"]
            json_path = os.path.join(self.project_path, "scienceai_ddb", tool["json_path"])
            if tool["tool_name"][:-19] not in files:
                files[tool["tool_name"][:-19]] = json_path
                tools[json_path] = csv_path
            else:
                if tool["tool_name"][-19:] > files[tool["tool_name"][:-19]][-19:]:
                    files[tool["tool_name"][:-19]] = json_path
                    tools[json_path] = csv_path
        return tools

    def get_analyst_tool_tracker(self, json_path):
        data = DDB.at(json_path).read()
        return data

    @log_update
    def clear_analyst_context(self, name):
        if not DDB.at(name).exists():
            raise ValueError(f"Analyst {name} not found")
        with DDB.at("Analysts").session() as (session, analysts):
            analysts[name]["contexts"] = []
            session.write()
        return True

    def get_analyst_metadata(self, name):
        if not DDB.at("Analysts", key=name).exists():
            raise ValueError(f"Analyst {name} not found")
        with DDB.at("Analysts").session() as (session, analysts):
            return analysts[name]

    def get_all_analysts(self):
        if not DDB.at("Analysts").exists():
            return []
        else:
            all = []
            for name in DDB.at("Analysts").read().keys():
                temp = self.get_analyst_metadata(name).copy()
                temp["name"] = name
                all.append(temp)
            return all

    def get_all_papers(self, analyst=None, named_list=None):
        if (analyst or named_list) and not (analyst and named_list):
            raise ValueError("Both analyst and named_list must be provided")
        papers = DDB.at("papers").read()
        if analyst:
            result = [paper for paper in papers.values() if analyst in paper and named_list in paper[analyst]]
            return result
        return list(papers.values())

    @log_update
    def remove_all_analyst_lists(self, analyst):
        with DDB.at("papers").session() as (session, papers):
            for paper_id in papers:
                self.remove_paper_from_list(paper_id, analyst)
        return True

    def get_paper_data(self, paper_id):
        if not DDB.at("papers", key=paper_id).exists():
            raise ValueError(f"Paper with id {paper_id} not found")
        paper = DDB.at("papers", key=paper_id).read()
        data = {}
        if DDB.at(paper_id).exists():
            data = DDB.at(paper_id).read()
        data["database"] = paper
        return data

    def get_all_papers_data(self, analyst=None, named_list=None, selected_key=None):
        papers = self.get_all_papers(analyst=analyst, named_list=named_list)
        output = []
        for paper in papers:
            paper_id = paper["paper_id"]
            paper_data = self.get_paper_data(paper_id)
            if selected_key:
                paper_data = paper_data.get(selected_key)
            else:
                output.append(paper_data)
        return output

    def get_analyst_data_visual(self, path):
        analysts = DDB.at("Analysts").read()
        if not analysts:
            return {}
        else:
            analysts = list(analysts.keys())
        if path == "/":
            return {k: {} for k in analysts}
        path_parts = path.strip("/").split("/")
        if len(path_parts) == 1:
            add_ons = {"evidence_files": {}, "internal_memory": {}}
            analyst = DDB.at(path_parts[0]).read()
            analyst.update(add_ons)
            return analyst
        if len(path_parts) == 2:
            metadata = self.get_analyst_metadata(path_parts[1]).copy()
            if not metadata:
                return {}
            metadata["evidence_files"] = {}
            metadata["internal_memory"] = {}
            if "tools" in metadata:
                del metadata["tools"]
            return metadata
        if len(path_parts) > 2 and path_parts[2] == "evidence_files":
            results = self.get_all_tool_trackers_for_analyst(path_parts[1])
            if len(path_parts) == 3:
                return {os.path.basename(tool)[:-25]: {} for tool in results.keys()}
            if len(path_parts) == 4:
                for json_path, csv_path in results.items():
                    if os.path.basename(json_path)[:-25] == path_parts[3]:
                        temp = {"json_path": json_path}
                        if csv_path:
                            temp["csv_path"] = csv_path
                        return temp
        if path_parts[2] == "internal_memory":
            results = self.get_analyst_context(path_parts[1], include_hidden=True)
            if len(path_parts) == 3:
                return {f"memory - {i+1}": {} for i in range(len(results))}
            if len(path_parts) > 3:
                result = results[int(path_parts[3].split(" - ")[1]) - 1]
                for part in path_parts[4:]:
                    result = result[part] if isinstance(result, dict) else result[int(part)]
                return result if isinstance(result, dict) else {str(i): v for i, v in enumerate(result)}

    def get_database_papers(self):
        full = DDB.at("papers").read()
        if not full:
            return []
        return [{**paper, **{"json_path": os.path.join(self.project_path, "scienceai_ddb", paper["paper_id"] + ".json")}} if "Title" in paper else paper for paper in full.values()]

    def get_database_chat(self):
        full = DDB.at("chat", key="messages").read()
        if not full:
            return []
        return full

    def get_last_message(self):
        if not DDB.at("chat", key="messages").exists():
            return None
        full = DDB.at("chat", key="messages").read()
        if len(full) == 0:
            return None
        return full[-1]

    def get_analyst_context(self, name, include_hidden=False):
        if not DDB.at(name).exists():
            raise ValueError(f"Analyst {name} not found")
        if include_hidden:
            return DDB.at(name).read()["context"]
        return [context for context in DDB.at(name).read()["context"] if not context.get("hidden", False)]

    def get_last_save(self, path=False):
        existing_save = None
        for project in os.listdir(self.storage_path):
            if project.find(self.project_name+"_-checkpoint-_") > -1:
                existing_save = project
        if path:
            if existing_save:
                return os.path.join(self.storage_path, existing_save)
        return existing_save

    def save_database(self):
        existing_save = self.get_last_save(path=True)
        new_save = self.project_name + "_-checkpoint-_" + self.update_time.replace(" ", "_").replace(":", "_")
        shutil.copytree(self.project_path, os.path.join(self.storage_path, new_save))
        if existing_save:
            shutil.rmtree(existing_save)